<?php
/**
Copyright Guillaume Bourgeois (2018)

contact@souke.fr

Ce logiciel est un programme informatique servant à aider les producteurs
à distribuer leur production en circuits courts.

Ce logiciel est régi par la licence CeCILL soumise au droit français et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL telle que diffusée par le CEA, le CNRS et l'INRIA
sur le site "http://www.cecill.info".

En contrepartie de l'accessibilité au code source et des droits de copie,
de modification et de redistribution accordés par cette licence, il n'est
offert aux utilisateurs qu'une garantie limitée.  Pour les mêmes raisons,
seule une responsabilité restreinte pèse sur l'auteur du programme,  le
titulaire des droits patrimoniaux et les concédants successifs.

A cet égard  l'attention de l'utilisateur est attirée sur les risques
associés au chargement,  à l'utilisation,  à la modification et/ou au
développement et à la reproduction du logiciel par l'utilisateur étant
donné sa spécificité de logiciel libre, qui peut le rendre complexe à
manipuler et qui le réserve donc à des développeurs et des professionnels
avertis possédant  des  connaissances  informatiques approfondies.  Les
utilisateurs sont donc invités à charger  et  tester  l'adéquation  du
logiciel à leurs besoins dans des conditions permettant d'assurer la
sécurité de leurs systèmes et ou de leurs données et, plus généralement,
à l'utiliser et l'exploiter dans les mêmes conditions de sécurité.

Le fait que vous puissiez accéder à cet en-tête signifie que vous avez
pris connaissance de la licence CeCILL, et que vous en avez accepté les
termes.
 */

namespace domain\Distribution\Distribution\Export;

use domain\Distribution\Distribution\Distribution;
use domain\Feature\Feature\Feature;
use domain\Feature\Feature\FeatureChecker;
use domain\Order\Order\Order;
use domain\Order\Order\OrderBuilder;
use domain\Order\Order\OrderRepository;
use domain\Order\Order\OrderSolver;
use domain\Producer\Producer\ProducerSolver;
use kartik\mpdf\Pdf;
use domain\_\AbstractGenerator;
use yii\base\ErrorException;
use yii\helpers\BaseStringHelper;

class DistributionShoppingCartLabelsPdfGenerator extends AbstractGenerator implements DistributionExportGeneratorInterface
{
    const FORMAT_70_42 = '70x42';
    const FORMAT_100_150 = '100x150';

    protected ProducerSolver $producerSolver;
    protected OrderRepository $orderRepository;
    protected OrderSolver $orderSolver;
    protected FeatureChecker $featureChecker;
    protected OrderBuilder $orderBuilder;

    public function loadDependencies(): void
    {
        $this->producerSolver = $this->loadService(ProducerSolver::class);
        $this->orderRepository = $this->loadService(OrderRepository::class);
        $this->orderSolver = $this->loadService(OrderSolver::class);
        $this->featureChecker = $this->loadService(FeatureChecker::class);
        $this->orderBuilder = $this->loadService(OrderBuilder::class);
    }

    public function getSpecificFormatDetailsArray(): array
    {
        return [
            self::FORMAT_70_42 => [70, 42],
            self::FORMAT_100_150 => [100, 150],
        ];
    }

    public function generate(Distribution $distribution, bool $save = false)
    {
        $isSpecificFormat = false;
        $exportShoppingCartLabelsFormat = $this->producerSolver->getConfig('export_shopping_cart_labels_format');
        $pdfInCartLabelSpecificFormat = (bool) $this->producerSolver->getConfig('export_shopping_cart_labels_pdf_in_cart_label_specific_format');
        $specificFormatWidth = 0;
        $specificFormatHeight = 0;
        if($this->featureChecker->isEnabled(Feature::ALIAS_EXPORT_SHOPPING_CART_LABELS_ADVANCED)) {
            if($exportShoppingCartLabelsFormat) {
                $isSpecificFormat = true;
                [$specificFormatWidth, $specificFormatHeight] = $this->getSpecificFormatDetails($exportShoppingCartLabelsFormat);
            }
            else {
                throw new ErrorException("Aucun format d'étiquette n'est défini dans les paramètres.");
            }
        }

        $ordersArray = $this->orderRepository->findOrdersByDistribution($distribution);
        $ordersArray = $this->filterOrdersExcludedUsersAndPointSales($ordersArray);

        foreach($ordersArray as $order) {
            $this->orderBuilder->initOrder($order);
        }

        $content = \Yii::$app->getView()->render('@backend/views/distribution/shopping-cart-labels.php', [
            'distribution' => $distribution,
            'ordersArray' => $ordersArray,
            'isSpecificFormat' => $isSpecificFormat,
            'specificFormatWidth' => $specificFormatWidth,
            'specificFormatHeight' => $specificFormatHeight,
            'shoppingCartLabelsPerColumn' => $this->producerSolver->getConfig('export_shopping_cart_labels_number_per_column') ?: 8,
            'pdfInCartLabelSpecificFormat' => $isSpecificFormat && $pdfInCartLabelSpecificFormat
        ]);

        if($isSpecificFormat) {
            $pdf = $this->getPdf($distribution, $content, $pdfInCartLabelSpecificFormat, true, $specificFormatWidth, $specificFormatHeight);
        }
        else {
            $pdf = $this->getPdf($distribution, $content, false, false);
            $pdf->getApi()->SetColumns(4);
            $pdf->getApi()->keepColumns = true;
        }

        return $pdf->render();
    }

    public function getPdf(Distribution $distribution, string $content, bool $pdfInCartLabelSpecificFormat, bool $isSpecificFormat, float $specificFormatWidth = 0, float $specificFormatHeight = 0)
    {
        if($pdfInCartLabelSpecificFormat) {
            $format = [$specificFormatWidth, $specificFormatHeight];
        }
        else {
            $format = Pdf::FORMAT_A4;
        }

        return new Pdf([
            'mode' => Pdf::MODE_UTF8,
            'format' => $format,
            'orientation' => Pdf::ORIENT_PORTRAIT,
            'destination' => Pdf::DEST_BROWSER,
            'filename' => \Yii::getAlias(
                '@app/web/pdf/Etiquettes-' . $distribution->date . '-' . $this->getProducerContextId() . '.pdf'
            ),
            'content' => $content,
            'cssInline' => $this->getCss($isSpecificFormat, $specificFormatWidth, $specificFormatHeight),
        ]);
    }

    public function filterOrdersExcludedUsersAndPointSales(array $ordersArray)
    {
        $ordersArrayFilter = [];

        foreach($ordersArray as $order) {
            if(!$this->isOrderUserExcludeExport($order)
                && !$this->isOrderPointSaleExcludeExport($order)) {
                $ordersArrayFilter[] = $order;
            }
        }

        return $ordersArrayFilter;
    }

    public function isOrderUserExcludeExport(Order $order): bool
    {
        return $order->user && $order->user->exclude_export_shopping_cart_labels;
    }

    public function isOrderPointSaleExcludeExport(Order $order): bool
    {
        return $order->pointSale && $order->pointSale->exclude_export_shopping_cart_labels;
    }

    public function getCss(bool $isSpecificFormat = false, float $specificFormatWidth = 0, float $specificFormatHeight = 0): string
    {
        $fontSizeArray = [
            'normal' => [
                'username' => 14,
                'phoneNumber' => 10,
                'pointSale' => 10,
                'comment' => 8,
                'products' => 10,
                'amount' => 10,
            ],
            'big' => [
                'username' => 18,
                'phoneNumber' => 14,
                'pointSale' => 14,
                'comment' => 12,
                'products' => 14,
                'amount' => 14,
            ]
        ];

        if($specificFormatWidth >= 100 && $specificFormatHeight >= 100) {
            $fontSizeArray = $fontSizeArray['big'];
        }
        else {
            $fontSizeArray = $fontSizeArray['normal'];
        }

        $css = '
            @page {
              margin: 0px 0px 0px 0px !important;
              padding: 0px 0px 0px 0px !important;
            }
            
            body {
              font-family: Arial, Helvetica, sans-serif;
            }
            
            .shopping-cart-label .username {
                font-weight: bold;
                font-size: '.$fontSizeArray['username'].'px;
            }
            
            .shopping-cart-label .phonenumber {
                font-size: '.$fontSizeArray['phoneNumber'].'px;
            }
            
            .shopping-cart-label .point-sale {
                font-size: '.$fontSizeArray['pointSale'].'px;
                text-transform: uppercase;
                margin-bottom: 3px;
            }
            
            .shopping-cart-label .comment {
                font-size: '.$fontSizeArray['comment'].'px;
                margin-bottom: 3px;
                font-style: italic;
            }
            
            .shopping-cart-label .products {
                 font-size: '.$fontSizeArray['products'].'px;
            }
            
            .shopping-cart-label .amount-and-payment {
                 font-size: '.$fontSizeArray['amount'].'px;
                 margin-top: 5px;
            }
            ';

        if($isSpecificFormat) {
            $paddingWidthShoppingCartLabel = 8;
            $paddingHeightShoppingCartLabel = 5;
            $specificFormatWidth = $specificFormatWidth - 2 * $paddingWidthShoppingCartLabel;
            $specificFormatHeight = $specificFormatHeight - 2 * $paddingHeightShoppingCartLabel + 1;

            $css .= '
            .shopping-cart-label {
                box-sizing: border-box;
                padding-top: '.$paddingHeightShoppingCartLabel.'mm;
                padding-bottom: '.$paddingHeightShoppingCartLabel.'mm;
                padding-left: '.$paddingWidthShoppingCartLabel.'mm;
                padding-right: '.$paddingWidthShoppingCartLabel.'mm;
                width: '.$specificFormatWidth.'mm;
                height: '.$specificFormatHeight.'mm;
                display: block;
                float: left;
            }
            
            .shopping-cart-label .producer-logo {
                width: 100px;
                height: 100px;
                
                padding-bottom: 20px;
            }
            
            .shopping-cart-label .amount-and-payment {
                
            }
            
            .shopping-cart-label .amount-and-payment .amount {
                display: inline-block; 
                background-color: #e0e0e0;
                color: #333;
                font-weight: bold;
                padding: 3px 8px;
                border-radius: 8px;
            }';
        }
        else {
            $css .= '
            .shopping-cart-label {
                -webkit-column-break-inside:avoid;
                column-break-inside: avoid;
                -webkit-page-break-inside:avoid;
                page-break-inside: avoid;
            }
            
            .shopping-cart-label .inner {
                   padding: 8px;
            }
            ';
        }

        return $css;
    }

    public function getShoppingCartLabelAsHtml(Order $order, int $index, bool $isSpecificFormat, int $specificFormatWidth, int $specificFormatHeight, bool $pdfInCartLabelSpecificFormat): string
    {
        $isSpecificFormatBig = $isSpecificFormat && $specificFormatWidth >= 100 && $specificFormatHeight >= 150;

        $amountAndPayment = '';
        if($isSpecificFormat) {
            $amountAndPayment = '<div class="amount-and-payment">';
            $amountAndPayment .= '<span class="amount">'.$this->orderSolver->getOrderAmountWithTax($order, Order::AMOUNT_TOTAL, true).'</span>';

            $paymentLabelPaid = $this->orderRepository->getPaymentLabelPaid($order);
            if($paymentLabelPaid && strlen($paymentLabelPaid)) {
                $amountAndPayment .= ' / '.$paymentLabelPaid;
            }
            else {
                $amountAndPayment .= ' / À régler';
            }
            $amountAndPayment .= '</div>';
        }

        $producerLogo = '';
        if($isSpecificFormatBig) {
            $producerLogo = '<div class="producer-logo"><img src="'.$this->producerSolver->getUrlLogo($this->getProducerContext()).'" /></div>';
        }

        $comment = '';
        if($isSpecificFormatBig && strlen($order->comment)) {
            $comment = '<div class="comment">« '.BaseStringHelper::truncate($order->comment, 50).' »</div>';
        }

        $phoneNumber = '';
        if($this->orderSolver->hasPhone($order)) {
            $phoneNumber = '<div class="phonenumber">'.$this->orderSolver->getPhone($order).'</div>';
        }

        return '<div class="shopping-cart-label shopping-cart-label-'.$index.'">
            <div class="inner">
                '.$producerLogo.'
                <div class="username">
                    '.$this->orderSolver->getOrderUsername($order).'
                </div>
                '.$phoneNumber.'
                <div class="point-sale">
                    '.date('d/m', strtotime($order->distribution->date)).' &bull;
                    '.$order->pointSale->name.'
                </div>
                '.$comment.'
                <div class="products">
                    '.$this->orderRepository->getCartSummary($order).'
                </div>
                '.$amountAndPayment.'
            </div>
        </div>';
    }

    public function getSpecificFormatDetails(string $name): array
    {
        $specificFormatDetailsArray = $this->getSpecificFormatDetailsArray();
        if(!isset($specificFormatDetailsArray[$name])) {
            throw new ErrorException("Format d'étiquette inconnu");
        }
        return $specificFormatDetailsArray[$name];
    }

    public function populateDropdownSpecificFormats(): array
    {
        $specificFormatsArray = [null => '--'];
        foreach($this->getSpecificFormatDetailsArray() as $key => $specificFormat) {
            $specificFormatsArray[$key] = $specificFormat[0].'x'.$specificFormat[1].' mm';
        }
        return $specificFormatsArray;
    }
}